@@ -31,7 +31,7 @@ end
 
 (** Provides a DNS-enabled {!Resolver_lwt} given a network stack. *)
 module Make
-    (R : Mirage_random.S)
+    (R : Mirage_crypto_rng_mirage.S)
     (T : Mirage_time.S)
     (C : Mirage_clock.MCLOCK)
     (P : Mirage_clock.PCLOCK)
